@@ -1,6 +1,6 @@
 {- git-annex repository initialization
  -
- - Copyright 2011-2022 Joey Hess <id@joeyh.name>
+ - Copyright 2011-2025 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -12,6 +12,7 @@ module Annex.Init (
 	checkInitializeAllowed,
 	ensureInitialized,
 	autoInitialize,
+	autoInitialize',
 	isInitialized,
 	initialize,
 	initialize',
@@ -73,17 +74,29 @@ data InitializeAllowed = InitializeAllowed
 
 checkInitializeAllowed :: (InitializeAllowed -> Annex a) -> Annex a
 checkInitializeAllowed a = guardSafeToUseRepo $ noAnnexFileContent' >>= \case
-	Nothing -> do
-		checkSqliteWorks
-		a InitializeAllowed
+	Nothing -> runAnnexHook' preInitAnnexHook annexPreInitCommand >>= \case
+		Nothing -> do
+			checkSqliteWorks
+			a InitializeAllowed
+		Just failedcommanddesc -> do
+			initpreventedby failedcommanddesc
+			notinitialized
 	Just noannexmsg -> do
-		warning "Initialization prevented by .noannex file (remove the file to override)"
+		initpreventedby ".noannex file (remove the file to override)"
 		unless (null noannexmsg) $
 			warning (UnquotedString noannexmsg)
-		giveup "Not initialized."
+		notinitialized
+  where
+	initpreventedby r = warning $ UnquotedString $
+		"Initialization prevented by " ++ r
+	notinitialized = giveup "Not initialized."
 
 initializeAllowed :: Annex Bool
-initializeAllowed = isNothing <$> noAnnexFileContent'
+initializeAllowed = noAnnexFileContent' >>= \case
+	Nothing -> runAnnexHook' preInitAnnexHook annexPreInitCommand >>= \case
+		Nothing -> return True
+		Just _ -> return False
+	Just _ -> return False
 
 noAnnexFileContent' :: Annex (Maybe String)
 noAnnexFileContent' = inRepo $
@@ -102,8 +115,8 @@ genDescription Nothing = do
 		Right username -> [username, at, hostname, ":", reldir]
 		Left _ -> [hostname, ":", reldir]
 
-initialize :: Maybe String -> Maybe RepoVersion -> Annex ()
-initialize mdescription mversion = checkInitializeAllowed $ \initallowed -> do
+initialize :: Annex () -> Maybe String -> Maybe RepoVersion -> Annex ()
+initialize startupannex mdescription mversion = checkInitializeAllowed $ \initallowed -> do
 	{- Has to come before any commits are made as the shared
 	 - clone heuristic expects no local objects. -}
 	sharedclone <- checkSharedClone
@@ -113,14 +126,14 @@ initialize mdescription mversion = checkInitializeAllowed $ \initallowed -> do
 	ensureCommit $ Annex.Branch.create
 
 	prepUUID
-	initialize' mversion initallowed
+	initialize' startupannex mversion initallowed
 	
 	initSharedClone sharedclone
 	
 	u <- getUUID
 	when (u == NoUUID) $
 		giveup "Failed to read annex.uuid from git config after setting it. This should never happen. Please file a bug report."
-	
+
 	{- Avoid overwriting existing description with a default
 	 - description. -}
 	whenM (pure (isJust mdescription) <||> not . M.member u <$> uuidDescMapRaw) $
@@ -128,8 +141,8 @@ initialize mdescription mversion = checkInitializeAllowed $ \initallowed -> do
 
 -- Everything except for uuid setup, shared clone setup, and initial
 -- description.
-initialize' :: Maybe RepoVersion -> InitializeAllowed -> Annex ()
-initialize' mversion _initallowed = do
+initialize' :: Annex () -> Maybe RepoVersion -> InitializeAllowed -> Annex ()
+initialize' startupannex mversion _initallowed = do
 	checkLockSupport
 	checkFifoSupport
 	checkCrippledFileSystem
@@ -161,6 +174,10 @@ initialize' mversion _initallowed = do
 	createInodeSentinalFile False
 	fixupUnusualReposAfterInit
 
+	-- This is usually run at Annex startup, but when git-annex was
+	-- not already initialized, it will not yet have run.
+	startupannex
+
 uninitialize :: Annex ()
 uninitialize = do
 	-- Remove hooks that are written when initializing.
@@ -202,12 +219,12 @@ getInitializedVersion = do
  -
  - Checks repository version and handles upgrades too.
  -}
-ensureInitialized :: Annex [Remote] -> Annex ()
-ensureInitialized remotelist = getInitializedVersion >>= maybe needsinit checkUpgrade
+ensureInitialized :: Annex () -> Annex [Remote] -> Annex ()
+ensureInitialized startupannex remotelist = getInitializedVersion >>= maybe needsinit checkUpgrade
   where
 	needsinit = ifM autoInitializeAllowed
 		( do
-			tryNonAsync (initialize Nothing Nothing) >>= \case
+			tryNonAsync (initialize startupannex Nothing Nothing) >>= \case
 				Right () -> noop
 				Left e -> giveup $ show e ++ "\n" ++
 					"git-annex: automatic initialization failed due to above problems"
@@ -255,15 +272,19 @@ guardSafeToUseRepo a = ifM (inRepo Git.Config.checkRepoConfigInaccessible)
  -
  - Checks repository version and handles upgrades too.
  -}
-autoInitialize :: Annex [Remote] -> Annex ()
-autoInitialize remotelist = getInitializedVersion >>= maybe needsinit checkUpgrade
+autoInitialize :: Annex () -> Annex [Remote] -> Annex ()
+autoInitialize = autoInitialize' autoInitializeAllowed
+
+autoInitialize' :: Annex Bool -> Annex () -> Annex [Remote] -> Annex ()
+autoInitialize' check startupannex remotelist =
+	getInitializedVersion >>= maybe needsinit checkUpgrade
   where
 	needsinit =
-		whenM (initializeAllowed <&&> autoInitializeAllowed) $ do
-			initialize Nothing Nothing
+		whenM (check <&&> initializeAllowed) $ do
+			initialize startupannex Nothing Nothing
 			autoEnableSpecialRemotes remotelist
 
-{- Checks if a repository is initialized. Does not check version for ugrade. -}
+{- Checks if a repository is initialized. Does not check version for upgrade. -}
 isInitialized :: Annex Bool
 isInitialized = maybe Annex.Branch.hasSibling (const $ return True) =<< getVersion
 
