@@ -1,6 +1,6 @@
 {- git-annex command
  -
- - Copyright 2013-2023 Joey Hess <id@joeyh.name>
+ - Copyright 2013-2024 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -17,6 +17,7 @@ import Text.Feed.Types
 import qualified Data.Set as S
 import qualified Data.Map as M
 import Data.Time.Clock
+import Data.Time.Clock.POSIX
 import Data.Time.Format
 import Data.Time.Calendar
 import Data.Time.LocalTime
@@ -37,7 +38,7 @@ import Logs.File
 import qualified Utility.Format
 import Utility.Tmp
 import Utility.Metered
-import Command.AddUrl (addUrlFile, downloadRemoteFile, parseDownloadOptions, DownloadOptions(..), checkCanAdd, addWorkTree, checkRaw)
+import Command.AddUrl (addUrlFile, downloadRemoteFile, parseDownloadOptions, DownloadOptions(..), checkClaimingUrl, checkCanAdd, addWorkTree, checkRaw, useYoutubeDl)
 import Annex.UUID
 import Backend.URL (fromUrl)
 import Annex.Content
@@ -48,9 +49,8 @@ import Logs.MetaData
 import Annex.MetaData
 import Annex.FileMatcher
 import Annex.UntrustedFilePath
-import qualified Annex.Branch
-import Logs
 import qualified Utility.RawFilePath as R
+import qualified Database.ImportFeed as Db
 
 cmd :: Command
 cmd = notBareRepo $ withAnnexOptions os $
@@ -62,6 +62,7 @@ cmd = notBareRepo $ withAnnexOptions os $
 data ImportFeedOptions = ImportFeedOptions
 	{ feedUrls :: CmdParams
 	, templateOption :: Maybe String
+	, scrapeOption :: Bool
 	, downloadOptions :: DownloadOptions
 	}
 
@@ -72,6 +73,10 @@ optParser desc = ImportFeedOptions
 		( long "template" <> metavar paramFormat
 		<> help "template for filenames"
 		))
+	<*> switch
+		( long "scrape"
+		<> help "scrape website for content to import"
+		)
 	<*> parseDownloadOptions False
 
 seek :: ImportFeedOptions -> CommandSeek
@@ -85,7 +90,7 @@ seek o = startConcurrency commandStages $ do
 		liftIO $ atomically $ do
 			m <- takeTMVar dlst
 			putTMVar dlst (M.insert url Nothing m)
-		commandAction $ getFeed url dlst
+		commandAction $ getFeed o url dlst
 		startpendingdownloads addunlockedmatcher cache dlst checkst False
 	
 	startpendingdownloads addunlockedmatcher cache dlst checkst True
@@ -136,18 +141,23 @@ seek o = startConcurrency commandStages $ do
 				clearFeedProblem url
 
 getFeed
-	:: URLString
+	:: ImportFeedOptions
+	-> URLString
 	-> TMVar (M.Map URLString (Maybe (Maybe [ToDownload])))
 	-> CommandStart
-getFeed url st =
+getFeed o url st =
 	starting "importfeed" (ActionItemOther (Just (UnquotedString url))) (SeekInput [url]) $
-		get `onException` recordfail
+		go `onException` recordfail
   where
 	record v = liftIO $ atomically $ do
 		m <- takeTMVar st
 		putTMVar st (M.insert url v m)
 	recordfail = record (Just Nothing)
 	
+	go
+		| scrapeOption o = scrape
+		| otherwise = get
+
 	get = withTmpFile "feed" $ \tmpf h -> do
 		liftIO $ hClose h
 		ifM (downloadFeed url tmpf)
@@ -160,7 +170,7 @@ getFeed url st =
 
 	-- Use parseFeedFromFile rather than reading the file
 	-- ourselves because it goes out of its way to handle encodings.
-	parse tmpf = liftIO (parseFeedFromFile' tmpf) >>= \case
+	parse tmpf = liftIO (parseFeedFromFile tmpf) >>= \case
 		Nothing -> debugfeedcontent tmpf "parsing the feed failed"
 		Just f -> do
 			case decodeBS $ fromFeedText $ getFeedTitle f of
@@ -182,19 +192,26 @@ getFeed url st =
 		recordfail
 		next $ feedProblem url
 			(msg ++ " (use --debug --debugfilter=ImportFeed to see the feed content that was downloaded)")
-
-parseFeedFromFile' :: FilePath -> IO (Maybe Feed)
-#if MIN_VERSION_feed(1,1,0)
-parseFeedFromFile' = parseFeedFromFile
-#else
-parseFeedFromFile' f = catchMaybeIO (parseFeedFromFile f)
-#endif
+		
+	scrape = youtubePlaylist url >>= \case
+		Left err -> do
+			recordfail
+			next $ feedProblem url err
+		Right playlist -> do
+			record (Just (Just (playlistDownloads url playlist)))
+			next $ return True
 
 data ToDownload = ToDownload
-	{ feed :: Feed
-	, feedurl :: URLString
-	, item :: Item
+	{ feedurl :: URLString
 	, location :: DownloadLocation
+	, itemid :: Maybe B.ByteString
+	-- Either the parsed or unparsed date.
+	, itempubdate :: Maybe (Either String UTCTime)
+	-- Fields that are used as metadata and to generate the filename.
+	, itemfields :: [(String, String)]
+	-- True when youtube-dl found this by scraping, so certainly
+	-- supports downloading it.
+	, youtubedlscraped :: Bool
 	}
 
 data DownloadLocation = Enclosure URLString | MediaLink URLString
@@ -202,65 +219,50 @@ data DownloadLocation = Enclosure URLString | MediaLink URLString
 type ItemId = String
 
 data Cache = Cache
-	{ knownurls :: S.Set URLString
-	, knownitems :: S.Set ItemId
+	{ dbhandle :: Maybe Db.ImportFeedDbHandle 
 	, template :: Utility.Format.Format
 	}
 
 getCache :: Maybe String -> Annex Cache
 getCache opttemplate = ifM (Annex.getRead Annex.force)
-	( ret S.empty S.empty
+	( ret Nothing
 	, do
 		j <- jsonOutputEnabled
 		unless j $
 			showStartMessage (StartMessage "importfeed" (ActionItemOther (Just "gathering known urls")) (SeekInput []))
-		(us, is) <- knownItems
+		h <- Db.openDb
 		unless j
 			showEndOk
-		ret (S.fromList us) (S.fromList is)
+		ret (Just h)
 	)
   where
 	tmpl = Utility.Format.gen $ fromMaybe defaultTemplate opttemplate
-	ret us is = return $ Cache us is tmpl
-
-{- Scan all url logs and metadata logs in the branch and find urls
- - and ItemIds that are already known. -}
-knownItems :: Annex ([URLString], [ItemId])
-knownItems = Annex.Branch.overBranchFileContents select (go [] []) >>= \case
-		Just r -> return r
-		Nothing -> giveup "This repository is read-only."
-  where
-	select f
-		| isUrlLog f = Just ()
-		| isMetaDataLog f = Just ()
-		| otherwise = Nothing
-
-	go uc ic reader = reader >>= \case
-		Just ((), f, Just content)
-			| isUrlLog f -> case parseUrlLog content of
-				[] -> go uc ic reader
-				us -> go (us++uc) ic reader
-			| isMetaDataLog f ->
-				let s = currentMetaDataValues itemIdField $
-					parseCurrentMetaData content
-				in if S.null s
-					then go uc ic reader
-					else go uc (map (decodeBS . fromMetaValue) (S.toList s)++ic) reader
-			| otherwise -> go uc ic reader
-		Just ((), _, Nothing) -> go uc ic reader
-		Nothing -> return (uc, ic)
+	ret h = return $ Cache h tmpl
 
 findDownloads :: URLString -> Feed -> [ToDownload]
 findDownloads u f = catMaybes $ map mk (feedItems f)
   where
 	mk i = case getItemEnclosure i of
 		Just (enclosureurl, _, _) ->
-			Just $ ToDownload f u i $ Enclosure $ 
-				decodeBS $ fromFeedText enclosureurl
+			Just $ mk' i
+				(Enclosure $ decodeBS $ fromFeedText enclosureurl)
 		Nothing -> case getItemLink i of
-			Just l -> Just $ ToDownload f u i $ 
-				MediaLink $ decodeBS $ fromFeedText l
+			Just l -> Just $ mk' i
+				(MediaLink $ decodeBS $ fromFeedText l)
 			Nothing -> Nothing
+	mk' i l = ToDownload
+		{ feedurl = u
+		, location = l
+		, itemid = case getItemId i of
+			Just (_, iid) -> Just (fromFeedText iid)
+			_ -> Nothing
+		, itempubdate = case getItemPublishDate i :: Maybe (Maybe UTCTime) of
+			Just (Just d) -> Just (Right d)
+			_ -> Left . decodeBS . fromFeedText 
+				<$> getItemPublishDateString  i
+		, itemfields = extractFeedItemFields f i u
+		, youtubedlscraped = False
+		}
 
 {- Feeds change, so a feed download cannot be resumed. -}
 downloadFeed :: URLString -> FilePath -> Annex Bool
@@ -274,7 +276,7 @@ startDownload addunlockedmatcher opts cache cv todownload = case location todown
 	Enclosure url -> startdownloadenclosure url
 	MediaLink linkurl -> do
 		let mediaurl = setDownloader linkurl YoutubeDownloader
-		let mediakey = Backend.URL.fromUrl mediaurl Nothing
+		let mediakey = Backend.URL.fromUrl mediaurl Nothing (verifiableOption (downloadOptions opts))
 		-- Old versions of git-annex that used quvi might have
 		-- used the quviurl for this, so check if it's known
 		-- to avoid adding it a second time.
@@ -285,14 +287,20 @@ startDownload addunlockedmatcher opts cache cv todownload = case location todown
 				, downloadmedia linkurl mediaurl mediakey
 				)
   where
-	forced = Annex.getRead Annex.force
-
 	{- Avoids downloading any items that are already known to be
-	 - associated with a file in the annex, unless forced. -}
-	checkknown url a
-		| knownitemid || S.member url (knownurls cache)
-			= ifM forced (a, nothingtodo)
-		| otherwise = a
+	 - associated with a file in the annex. -}
+	checkknown url a = case dbhandle cache of
+		Just db -> ifM (liftIO $ Db.isKnownUrl db url)
+			( nothingtodo
+			, case itemid todownload of
+				Just iid ->
+					ifM (liftIO $ Db.isKnownItemId db iid)
+						( nothingtodo
+						, a
+						)
+				Nothing -> a
+			)
+		Nothing -> a
 
 	nothingtodo = recordsuccess >> stop
 
@@ -302,14 +310,8 @@ startDownload addunlockedmatcher opts cache cv todownload = case location todown
 	startdownloadenclosure url = checkknown url $ startUrlDownload cv todownload url $
 		downloadEnclosure addunlockedmatcher opts cache cv todownload url 
 
-	knownitemid = case getItemId (item todownload) of
-		Just (_, itemid) ->
-			S.member (decodeBS $ fromFeedText itemid) (knownitems cache)
-		_ -> False
-
-	downloadmedia linkurl mediaurl mediakey
-		| rawOption (downloadOptions opts) = startdownloadlink
-		| otherwise = ifM (youtubeDlSupported linkurl)
+	downloadmedia linkurl mediaurl mediakey =
+		ifM (useYoutubeDl (downloadOptions opts) <&&> youtubeDlSupported linkurl)
 			( startUrlDownload cv todownload linkurl $
 				withTmpWorkDir mediakey $ \workdir -> do
 					dl <- youtubeDl linkurl (fromRawFilePath workdir) nullMeterUpdate
@@ -338,8 +340,8 @@ startDownload addunlockedmatcher opts cache cv todownload = case location todown
 		contdownloadlink = downloadEnclosure addunlockedmatcher opts cache cv todownload linkurl
 
 	addmediafast linkurl mediaurl mediakey =
-		ifM (pure (not (rawOption (downloadOptions opts)))
-		     <&&> youtubeDlSupported linkurl)
+		ifM (useYoutubeDl (downloadOptions opts)
+		     <&&> (pure (youtubedlscraped todownload) <||> youtubeDlSupported linkurl))
 			( startUrlDownload cv todownload linkurl $ do
 				runDownload todownload linkurl ".m" cache cv $ \f ->
 					checkCanAdd (downloadOptions opts) f $ \canadd -> do
@@ -352,7 +354,7 @@ downloadEnclosure :: AddUnlockedMatcher -> ImportFeedOptions -> Cache -> TMVar B
 downloadEnclosure addunlockedmatcher opts cache cv todownload url = 
 	runDownload todownload url (takeWhile (/= '?') $ takeExtension url) cache cv $ \f -> do
 		let f' = fromRawFilePath f
-		r <- Remote.claimingUrl url
+		r <- checkClaimingUrl (downloadOptions opts) url
 		if Remote.uuid r == webUUID || rawOption (downloadOptions opts)
 			then checkRaw (Just url) (downloadOptions opts) (pure Nothing) $ do
 				let dlopts = (downloadOptions opts)
@@ -402,7 +404,7 @@ runDownload todownload url extension cache cv getter = do
 	case dest of
 		Nothing -> do
 			recordsuccess
-			stop
+			next $ return True
 		Just f -> getter (toRawFilePath f) >>= \case
 			Just ks
 				-- Download problem.
@@ -452,7 +454,7 @@ runDownload todownload url extension cache cv getter = do
 				in d </> show n ++ "_" ++ base
 		tryanother = makeunique (n + 1) file
 		alreadyexists = liftIO $ isJust <$> catchMaybeIO (R.getSymbolicLinkStatus (toRawFilePath f))
-		checksameurl k = ifM (elem url <$> getUrls k)
+		checksameurl k = ifM (elem url . map fst . map getDownloader <$> getUrls k)
 			( return Nothing
 			, tryanother
 			)
@@ -479,9 +481,9 @@ defaultTemplate = "${feedtitle}/${itemtitle}${extension}"
 feedFile :: Utility.Format.Format -> ToDownload -> String -> FilePath
 feedFile tmpl i extension = sanitizeLeadingFilePathCharacter $ 
 	Utility.Format.format tmpl $
-		M.map sanitizeFilePathComponent $ M.fromList $ extractFields i ++
+		M.map sanitizeFilePathComponent $ M.fromList $ itemfields i ++
 			[ ("extension", extension)
-			, extractField "itempubdate" [itempubdate]
+			, extractField "itempubdate" [itempubdatestring]
 			, extractField "itempubyear" [itempubyear]
 			, extractField "itempubmonth" [itempubmonth]
 			, extractField "itempubday" [itempubday]
@@ -490,18 +492,13 @@ feedFile tmpl i extension = sanitizeLeadingFilePathCharacter $
 			, extractField "itempubsecond" [itempubsecond]
 			]
   where
-	itm = item i
-
-	pubdate = case getItemPublishDate itm :: Maybe (Maybe UTCTime) of
-		Just (Just d) -> Just d
-		_ -> Nothing
+	pubdate = maybe Nothing eitherToMaybe (itempubdate i)
 	
-	itempubdate = case pubdate of
-		Just pd -> Just $
-			formatTime defaultTimeLocale "%F" pd
+	itempubdatestring = case itempubdate i of
+		Just (Right pd) -> Just $ formatTime defaultTimeLocale "%F" pd
 		-- if date cannot be parsed, use the raw string
-		Nothing-> replace "/" "-" . decodeBS . fromFeedText
-			<$> getItemPublishDateString itm
+		Just (Left s) -> Just $ replace "/" "-" s
+		Nothing -> Nothing
 	
 	(itempubyear, itempubmonth, itempubday) = case pubdate of
 		Nothing -> (Nothing, Nothing, Nothing)
@@ -520,43 +517,21 @@ feedFile tmpl i extension = sanitizeLeadingFilePathCharacter $
 			   )
 
 extractMetaData :: ToDownload -> MetaData
-extractMetaData i = case getItemPublishDate (item i) :: Maybe (Maybe UTCTime) of
-	Just (Just d) -> unionMetaData meta (dateMetaData d meta)
+extractMetaData i = case itempubdate i of
+	Just (Right d) -> unionMetaData meta (dateMetaData d meta)
 	_ -> meta
   where
 	tometa (k, v) = (mkMetaFieldUnchecked (T.pack k), S.singleton (toMetaValue (encodeBS v)))
-	meta = MetaData $ M.fromList $ map tometa $ extractFields i
+	meta = MetaData $ M.fromList $ map tometa $ itemfields i
 
 minimalMetaData :: ToDownload -> MetaData
-minimalMetaData i = case getItemId (item i) of
-	(Nothing) -> emptyMetaData
-	(Just (_, itemid)) -> MetaData $ M.singleton itemIdField 
-		(S.singleton $ toMetaValue $ fromFeedText itemid)
+minimalMetaData i = case itemid i of
+	Nothing -> emptyMetaData
+	Just iid -> MetaData $ M.singleton itemIdField
+		(S.singleton $ toMetaValue iid)
 
-{- Extract fields from the feed and item, that are both used as metadata,
- - and to generate the filename. -}
-extractFields :: ToDownload -> [(String, String)]
-extractFields i = map (uncurry extractField)
-	[ ("feedurl", [Just (feedurl i)])
-	, ("feedtitle", [feedtitle])
-	, ("itemtitle", [itemtitle])
-	, ("feedauthor", [feedauthor])
-	, ("itemauthor", [itemauthor])
-	, ("itemsummary", [decodeBS . fromFeedText <$> getItemSummary (item i)])
-	, ("itemdescription", [decodeBS . fromFeedText <$> getItemDescription (item i)])
-	, ("itemrights", [decodeBS . fromFeedText <$> getItemRights (item i)])
-	, ("itemid", [decodeBS . fromFeedText . snd <$> getItemId (item i)])
-	, ("title", [itemtitle, feedtitle])
-	, ("author", [itemauthor, feedauthor])
-	]
-  where
-	feedtitle = Just $ decodeBS $ fromFeedText $ getFeedTitle $ feed i
-	itemtitle = decodeBS . fromFeedText <$> getItemTitle (item i)
-	feedauthor = decodeBS . fromFeedText <$> getFeedAuthor (feed i)
-	itemauthor = decodeBS . fromFeedText <$> getItemAuthor (item i)
-
-itemIdField :: MetaField
-itemIdField = mkMetaFieldUnchecked "itemid"
+noneValue :: String
+noneValue = "none"
 
 extractField :: String -> [Maybe String] -> (String, String)
 extractField k [] = (k, noneValue)
@@ -564,8 +539,56 @@ extractField k (Just v:_)
 	| not (null v) = (k, v)
 extractField k (_:rest) = extractField k rest
 
-noneValue :: String
-noneValue = "none"
+extractFeedItemFields :: Feed -> Item -> URLString -> [(String, String)]
+extractFeedItemFields f i u = map (uncurry extractField)
+	[ ("feedurl", [Just u])
+	, ("feedtitle", [feedtitle])
+	, ("itemtitle", [itemtitle])
+	, ("feedauthor", [feedauthor])
+	, ("itemauthor", [itemauthor])
+	, ("itemsummary", [decodeBS . fromFeedText <$> getItemSummary i])
+	, ("itemdescription", [decodeBS . fromFeedText <$> getItemDescription i])
+	, ("itemrights", [decodeBS . fromFeedText <$> getItemRights i])
+	, ("itemid", [decodeBS . fromFeedText . snd <$> getItemId i])
+	, ("title", [itemtitle, feedtitle])
+	, ("author", [itemauthor, feedauthor])
+	]
+  where
+	feedtitle = Just $ decodeBS $ fromFeedText $ getFeedTitle f
+	itemtitle = decodeBS . fromFeedText <$> getItemTitle i
+	feedauthor = decodeBS . fromFeedText <$> getFeedAuthor f
+	itemauthor = decodeBS . fromFeedText <$> getItemAuthor i
+
+playlistFields :: URLString -> YoutubePlaylistItem -> [(String, String)]
+playlistFields u i = map (uncurry extractField)
+	[ ("feedurl", [Just u])
+	, ("feedtitle", [youtube_playlist_title i])
+	, ("itemtitle", [youtube_title i])
+	, ("feedauthor", [youtube_playlist_uploader i])
+	, ("itemauthor", [youtube_playlist_uploader i])
+	-- itemsummary omitted, no equivalent in yt-dlp data
+	, ("itemdescription", [youtube_description i])
+	, ("itemrights", [youtube_license i])
+	, ("itemid", [youtube_url i])
+	, ("title", [youtube_title i, youtube_playlist_title i])
+	, ("author", [youtube_playlist_uploader i])
+	]
+
+playlistDownloads :: URLString -> [YoutubePlaylistItem] -> [ToDownload]
+playlistDownloads url = mapMaybe go
+  where
+	go i = do
+		iurl <- youtube_url i
+		return $ ToDownload
+			{ feedurl = url
+			, location = MediaLink iurl
+			, itemid = Just (encodeBS iurl)
+			, itempubdate = 
+				Right . posixSecondsToUTCTime . fromIntegral
+					<$> youtube_timestamp i
+			, itemfields = playlistFields url i
+			, youtubedlscraped = True
+			}
 
 {- Called when there is a problem with a feed.
  -
@@ -608,7 +631,7 @@ clearFeedProblem url =
 		=<< feedState url
 
 feedState :: URLString -> Annex RawFilePath
-feedState url = fromRepo $ gitAnnexFeedState $ fromUrl url Nothing
+feedState url = fromRepo $ gitAnnexFeedState $ fromUrl url Nothing False
 
 {- The feed library parses the feed to Text, and does not use the
  - filesystem encoding to do it, so when the locale is not unicode
