@@ -1,7 +1,7 @@
 {- git-annex command
  -
  - Copyright 2011 Joachim Breitner <mail@joachim-breitner.de>
- - Copyright 2011-2023 Joey Hess <id@joeyh.name>
+ - Copyright 2011-2024 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
@@ -19,7 +19,6 @@ module Command.Sync (
 	prepMerge,
 	mergeLocal,
 	mergeRemote,
-	commitStaged,
 	commitMsg,
 	pushBranch,
 	updateBranch,
@@ -29,6 +28,7 @@ module Command.Sync (
 	parseUnrelatedHistoriesOption,
 	SyncOptions(..),
 	OperationMode(..),
+	syncBranch,
 ) where
 
 import Command
@@ -36,7 +36,6 @@ import qualified Annex
 import qualified Annex.Branch
 import qualified Remote
 import qualified Types.Remote as Remote
-import Annex.Hook
 import qualified Git.Command
 import qualified Git.LsFiles as LsFiles
 import qualified Git.Branch
@@ -54,10 +53,12 @@ import Annex.Path
 import Annex.Wanted
 import Annex.Content
 import Annex.WorkTree
+import Annex.FileMatcher
 import Command.Get (getKey')
 import qualified Command.Move
 import qualified Command.Export
 import qualified Command.Import
+import qualified Command.Migrate
 import Annex.Drop
 import Annex.UUID
 import Logs.UUID
@@ -76,18 +77,18 @@ import Annex.TaggedPush
 import Annex.CurrentBranch
 import Annex.Import
 import Annex.CheckIgnore
-import Types.FileMatcher
+import Annex.PidLock
 import Types.GitConfig
+import Types.Availability
 import qualified Database.Export as Export
 import Utility.Bloom
 import Utility.OptParse
 import Utility.Process.Transcript
 import Utility.Tuple
+import Utility.Matcher
 
 import Control.Concurrent.MVar
 import qualified Data.Map as M
-import qualified Data.ByteString as S
-import Data.Char
 
 cmd :: Command
 cmd = withAnnexOptions [jobsOption, backendOption] $
@@ -104,7 +105,7 @@ data SyncOptions = SyncOptions
 	, notOnlyAnnexOption :: Bool
 	, commitOption :: Bool
 	, noCommitOption :: Bool
-	, messageOption :: Maybe String
+	, messageOption :: [String]
 	, pullOption :: Bool
 	, pushOption :: Bool
 	, contentOption :: Maybe Bool
@@ -124,7 +125,7 @@ instance Default SyncOptions where
 		, notOnlyAnnexOption = False
 		, commitOption = False
 		, noCommitOption = False
-		, messageOption = Nothing
+		, messageOption = []
 		, pullOption = False
 		, pushOption = False
 		, contentOption = Just False
@@ -168,8 +169,8 @@ optParser mode desc = SyncOptions
 			( long "no-commit"
 			<> help "avoid git commit" 
 			))
-	<*> unlessmode [SyncMode, AssistMode] Nothing
-		(optional (strOption
+	<*> unlessmode [SyncMode, AssistMode] []
+		(many (strOption
 			( long "message" <> short 'm' <> metavar "MSG"
 			<> help "commit message"
 			)))
@@ -256,8 +257,6 @@ instance DeferredParseClass SyncOptions where
 
 seek :: SyncOptions -> CommandSeek
 seek o = do
-	warnSyncContentTransition o
-	
 	prepMerge
 	
 	seek' o
@@ -267,8 +266,9 @@ seek' o = startConcurrency transferStages $ do
 	let withbranch a = a =<< getCurrentBranch
 
 	remotes <- syncRemotes (syncWith o)
-	-- Remotes that are git repositories, not (necesarily) special remotes.
-	let gitremotes = filter (Remote.gitSyncableRemoteType . Remote.remotetype) remotes
+	warnSyncContentTransition o remotes
+	-- Remotes that git can push to and pull from.
+	let gitremotes = filter Remote.gitSyncableRemote remotes
 	-- Remotes that contain annex object content.
 	contentremotes <- filter (\r -> Remote.uuid r /= NoUUID)
 		<$> filterM (not <$$> liftIO . getDynamicConfig . remoteAnnexIgnore . Remote.gitconfig) remotes
@@ -293,6 +293,10 @@ seek' o = startConcurrency transferStages $ do
 			
 			content <- shouldSyncContent o
 
+			when content $
+				whenM (annexSyncMigrations <$> Annex.getGitConfig) $
+					Command.Migrate.seekDistributedMigrations True
+
 			forM_ (filter isImport contentremotes) $
 				withbranch . importRemote content o
 			forM_ (filter isThirdPartyPopulated contentremotes) $
@@ -303,15 +307,15 @@ seek' o = startConcurrency transferStages $ do
 				-- repositories, in case that lets content
 				-- be dropped from other repositories.
 				exportedcontent <- withbranch $
-					seekExportContent (Just o)
-						(filter isExport contentremotes)
+					seekExportContent (Just o) contentremotes
 
 				-- Sync content with remotes, including
 				-- importing from import remotes (since
 				-- importing only downloads new files not
 				-- old files)
 				let shouldsynccontent r
-					| isExport r && not (isImport r) = False
+					| isExport r && not (isImport r) 
+						&& not (exportHasAnnexObjects r) = False
 					| otherwise = True
 				syncedcontent <- withbranch $
 					seekSyncContent o
@@ -352,15 +356,16 @@ mergeConfig mergeunrelated = do
 		]
 
 merge :: CurrBranch -> [Git.Merge.MergeConfig] -> SyncOptions -> Git.Branch.CommitMode -> [Git.Branch] -> Annex Bool
-merge currbranch mergeconfig o commitmode tomergel = do
-	canresolvemerge <- if resolveMergeOverride o
-		then getGitConfigVal annexResolveMerge
-		else return False
-	and <$> case currbranch of
-		(Just b, Just adj) -> forM tomergel $ \tomerge ->
-			mergeToAdjustedBranch tomerge (b, adj) mergeconfig canresolvemerge commitmode
-		(b, _) -> forM tomergel $ \tomerge ->
-			autoMergeFrom tomerge b mergeconfig commitmode canresolvemerge
+merge currbranch mergeconfig o commitmode tomergel = 
+	runsGitAnnexChildProcessViaGit $ do
+		canresolvemerge <- if resolveMergeOverride o
+			then getGitConfigVal annexResolveMerge
+			else return False
+		and <$> case currbranch of
+			(Just b, Just adj) -> forM tomergel $ \tomerge ->
+				mergeToAdjustedBranch tomerge (b, adj) mergeconfig canresolvemerge commitmode
+			(b, _) -> forM tomergel $ \tomerge ->
+				autoMergeFrom tomerge b mergeconfig commitmode canresolvemerge
 
 syncBranch :: Git.Branch -> Git.Branch
 syncBranch = Git.Ref.underBase "refs/heads/synced" . origBranch
@@ -371,10 +376,10 @@ origBranch = fromViewBranch . fromAdjustedBranch
 remoteBranch :: Remote -> Git.Ref -> Git.Ref
 remoteBranch remote = Git.Ref.underBase $ "refs/remotes/" ++ Remote.name remote
 
--- Do automatic initialization of remotes when possible when getting remote
--- list.
 syncRemotes :: [String] -> Annex [Remote]
 syncRemotes ps = do
+	-- Do automatic initialization of remotes when possible
+	-- when getting remote list.
 	remotelist <- Remote.remoteList' True
 	available <- filterM (liftIO . getDynamicConfig . remoteAnnexSync . Remote.gitconfig) remotelist
 	syncRemotes' ps available
@@ -389,26 +394,26 @@ syncRemotes' ps available =
 	
 	listed = concat <$> mapM Remote.byNameOrGroup ps
 	
-	good r
-		| Remote.gitSyncableRemoteType (Remote.remotetype r) =
-			Remote.Git.repoAvail =<< Remote.getRepo r
-		| otherwise = return True
+	good r = tryNonAsync (Remote.availability r) >>= return . \case
+		Right Unavailable -> False
+		_ -> True
 	
 	fastest = fromMaybe [] . headMaybe . Remote.byCost
 
 commit :: SyncOptions -> CommandStart
 commit o = stopUnless shouldcommit $ starting "commit" ai si $ do
-	commitmessage <- maybe commitMsg return (messageOption o)
 	Annex.Branch.commit =<< Annex.Branch.commitMessage
+	mopts <- concatMap (\msg -> [Param "-m", Param msg])
+		<$> if null (messageOption o)
+			then (:[]) <$> commitMsg
+			else pure (messageOption o)
 	next $ do
 		showOutput
 		let cmode = Git.Branch.ManualCommit
 		cquiet <- Git.Branch.CommitQuiet <$> commandProgressDisabled
-		void $ inRepo $ Git.Branch.commitCommand cmode cquiet
-			[ Param "-a"
-			, Param "-m"
-			, Param commitmessage
-			]
+		void $ inRepo $ Git.Branch.commitCommand
+			cmode cquiet
+			([ Param "-a" ] ++ mopts)
 		return True
   where
 	shouldcommit = notOnlyAnnex o <&&>
@@ -422,18 +427,8 @@ commitMsg :: Annex String
 commitMsg = do
 	u <- getUUID
 	m <- uuidDescMap
-	return $ "git-annex in " ++ maybe "unknown" fromUUIDDesc (M.lookup u m)
-
-commitStaged :: Git.Branch.CommitMode -> String -> Annex Bool
-commitStaged commitmode commitmessage = do
-	runAnnexHook preCommitAnnexHook
-	mb <- inRepo Git.Branch.currentUnsafe
-	let (getparent, branch) = case mb of
-		Just b -> (Git.Ref.sha b, b)
-		Nothing -> (Git.Ref.headSha, Git.Ref.headRef)
-	parents <- maybeToList <$> inRepo getparent
-	void $ inRepo $ Git.Branch.commit commitmode False commitmessage branch parents
-	return True
+	return $ "git-annex in "
+		++ maybe "unknown" fromUUIDDesc (M.lookup u m)
 
 mergeLocal :: [Git.Merge.MergeConfig] -> SyncOptions -> CurrBranch -> CommandStart
 mergeLocal mergeconfig o currbranch = stopUnless (notOnlyAnnex o) $
@@ -585,7 +580,8 @@ importRemote importcontent o remote currbranch
 			let (branch, subdir) = splitRemoteAnnexTrackingBranchSubdir b
 			if canImportKeys remote importcontent
 				then do
-					Command.Import.seekRemote remote branch subdir importcontent (CheckGitIgnore True)
+					addunlockedmatcher <- addUnlockedMatcher
+					Command.Import.seekRemote remote branch subdir importcontent (CheckGitIgnore True) addunlockedmatcher []
 					-- Importing generates a branch
 					-- that is not initially connected
 					-- to the current branch, so allow
@@ -612,7 +608,7 @@ pullThirdPartyPopulated o remote
   where
 	go (Just importable) = importChanges remote ImportTree False True importable >>= \case
 		ImportFinished imported -> do
-			(_t, updatestate) <- recordImportTree remote ImportTree imported
+			(_t, updatestate) <- recordImportTree remote ImportTree Nothing imported
 			next $ do
 				updatestate
 				return True
@@ -841,7 +837,7 @@ seekSyncContent o rs currbranch = do
   where
 	seekworktree mvar l bloomfeeder = do
 		let seeker = AnnexedFileSeeker
-			{ startAction = gofile bloomfeeder mvar
+			{ startAction = const $ gofile bloomfeeder mvar
 			, checkContentPresent = Nothing
 			, usesLocationLog = True
 			}
@@ -898,7 +894,7 @@ seekSyncContent o rs currbranch = do
 syncFile :: SyncOptions -> Either (Maybe (Bloom Key)) (Key -> Annex ()) -> [Remote] -> AssociatedFile -> Key -> Annex Bool
 syncFile o ebloom rs af k = do
 	inhere <- inAnnex k
-	locs <- map Remote.uuid <$> Remote.keyPossibilities k
+	locs <- map Remote.uuid <$> Remote.keyPossibilities (Remote.IncludeIgnored False) k
 	let (have, lack) = partition (\r -> Remote.uuid r `elem` locs) rs
 
 	got <- anyM id =<< handleget have inhere
@@ -932,37 +928,41 @@ syncFile o ebloom rs af k = do
 	
 	return (got || not (null putrs))
   where
-	wantget have inhere = allM id 
+	wantget lu have inhere = allM id 
 		[ pure (pullOption o || operationMode o == SatisfyMode)
 		, pure (not $ null have)
 		, pure (not inhere)
-		, wantGet True (Just k) af
+		, wantGet lu True (Just k) af
 		]
-	handleget have inhere = ifM (wantget have inhere)
-		( return [ get have ]
-		, return []
-		)
-	get have = includeCommandAction $ starting "get" ai si $
-		stopUnless (getKey' k af have) $
+	handleget have inhere = do
+		lu <- prepareLiveUpdate Nothing k AddingKey
+		ifM (wantget lu have inhere)
+			( return [ get lu have ]
+			, return []
+			)
+	get lu have = includeCommandAction $ starting "get" ai si $
+		stopUnless (getKey' lu k af have) $
 			next $ return True
 
-	wantput r
+	wantput lu r
 		| pushOption o == False && operationMode o /= SatisfyMode = return False
 		| Remote.readonly r || remoteAnnexReadOnly (Remote.gitconfig r) = return False
-		| isExport r = return False
+		| isImport r && not (isExport r) = return False
+		| isExport r && not (exportHasAnnexObjects r) = return False
 		| isThirdPartyPopulated r = return False
-		| otherwise = wantGetBy True (Just k) af (Remote.uuid r)
+		| otherwise = wantGetBy lu True (Just k) af (Remote.uuid r)
 	handleput lack inhere
 		| inhere = catMaybes <$>
-			( forM lack $ \r ->
-				ifM (wantput r <&&> put r)
+			( forM lack $ \r -> do
+				lu <- prepareLiveUpdate (Just (Remote.uuid r)) k AddingKey
+				ifM (wantput lu r <&&> put lu r)
 					( return (Just (Remote.uuid r))
 					, return Nothing
 					)
 			)
 		| otherwise = return []
-	put dest = includeCommandAction $ 
-		Command.Move.toStart' dest Command.Move.RemoveNever af k ai si
+	put lu dest = includeCommandAction $ 
+		Command.Move.toStart' lu dest Command.Move.RemoveNever af k ai si
 
 	ai = mkActionItem (k, af)
 	si = SeekInput []
@@ -980,14 +980,30 @@ syncFile o ebloom rs af k = do
  - Returns True if any file transfers were made.
  -}
 seekExportContent :: Maybe SyncOptions -> [Remote] -> CurrBranch -> Annex Bool
-seekExportContent o rs (currbranch, _) = or <$> forM rs go
+seekExportContent o rs currbranch =
+	seekExportContent' o (filter canexportcontent rs) currbranch
+  where
+	canexportcontent r = isExport r && not (isProxied r)
+
+seekExportContent' :: Maybe SyncOptions -> [Remote] -> CurrBranch -> Annex Bool
+seekExportContent' o rs (mcurrbranch, madj)
+	| null rs = return False
+	| otherwise = do
+		-- Propagate commits from the adjusted branch, so that
+		-- when the remoteAnnexTrackingBranch is set to the parent
+		-- branch, it will be up-to-date.
+		case (mcurrbranch, madj) of
+			(Just currbranch, Just adj) ->
+				propigateAdjustedCommits currbranch adj
+			_ -> noop
+		or <$> forM rs go
   where
 	go r
 		| maybe False (\o' -> operationMode o' == SatisfyMode) o =
 			case remoteAnnexTrackingBranch (Remote.gitconfig r) of
 				Nothing -> return False
-				Just b -> withdb r $ \db ->
-					cannotupdateexport r db (Just b) False
+				Just _ -> withdb r $ \db ->
+					cannotupdateexport r db Nothing False
 		| not (maybe True pushOption o) = return False
 		| not (remoteAnnexPush (Remote.gitconfig r)) = return False
 		| otherwise = withdb r (go' r)
@@ -1001,29 +1017,31 @@ seekExportContent o rs (currbranch, _) = or <$> forM rs go
 				Nothing -> id
 			mcurrtree <- maybe (pure Nothing)
 				(inRepo . Git.Ref.tree . addsubdir)
-				currbranch
+				mcurrbranch
 			mtbcommitsha <- Command.Export.getExportCommit r b
 			case (mtree, mcurrtree, mtbcommitsha) of
 				(Just tree, Just currtree, Just _)
 					| tree == currtree -> do
 						filteredtree <- Command.Export.filterExport r tree
 						Command.Export.changeExport r db filteredtree
-						Command.Export.fillExport r db filteredtree mtbcommitsha
-					| otherwise -> cannotupdateexport r db (Just b) False
-				_ -> cannotupdateexport r db (Just b) True
+						Command.Export.fillExport r db filteredtree mtbcommitsha []
+					| otherwise -> cannotupdateexport r db Nothing False
+				(Nothing, _, _) -> cannotupdateexport r db (Just (Git.fromRef b ++ " does not exist")) True
+				(_, Nothing, _) -> cannotupdateexport r db (Just "no branch is currently checked out") True
+				(_, _, Nothing) -> cannotupdateexport r db (Just "tracking branch name is not valid") True
 	
 	withdb r a = bracket
 		(Export.openDb (Remote.uuid r))
 		Export.closeDb
 		(\db -> Export.writeLockDbWhile db (a db))
 	
-	cannotupdateexport r db mtb showwarning = do
+	cannotupdateexport r db mreason showwarning = do
 		exported <- getExport (Remote.uuid r)
 		when showwarning $
-			maybe noop (warncannotupdateexport r mtb exported) currbranch
+			maybe noop (warncannotupdateexport r mreason exported) mcurrbranch
 		fillexistingexport r db (exportedTreeishes exported) Nothing
 	
-	warncannotupdateexport r mtb exported currb = case mtb of
+	warncannotupdateexport r mreason exported currb = case mreason of
 		Nothing -> inRepo (Git.Ref.tree currb) >>= \case
 			Just currt | not (any (== currt) (exportedTreeishes exported)) ->
 				showLongNote $ UnquotedString $ unwords
@@ -1033,9 +1051,9 @@ seekExportContent o rs (currbranch, _) = or <$> forM rs go
 					, "(Set " ++ gitconfig ++ " to enable it.)"
 					]
 			_ -> noop
-		Just b -> showLongNote $ UnquotedString $ unwords
+		Just reason -> showLongNote $ UnquotedString $ unwords
 			[ notupdating
-			, "because " ++ Git.fromRef b ++ " does not exist."
+			, "because " ++ reason ++ "."
 			, "(As configured by " ++ gitconfig ++ ")"
 			]
 	  where
@@ -1049,7 +1067,7 @@ seekExportContent o rs (currbranch, _) = or <$> forM rs go
 		-- filling in any files that did not get transferred
 		-- to the existing exported tree.
 		let filteredtree = Command.Export.ExportFiltered tree
-		Command.Export.fillExport r db filteredtree mtbcommitsha
+		Command.Export.fillExport r db filteredtree mtbcommitsha []
 	fillexistingexport r _ _ _ = do
 		warnExportImportConflict r
 		return False
@@ -1102,23 +1120,30 @@ shouldSyncContent o
 			HasGitConfig (Just c) -> return c
 			_ -> return d
 
--- Transition started May 2023, should wait until that has been in a Debian
--- stable release before completing the transition.
-warnSyncContentTransition :: SyncOptions -> Annex ()
-warnSyncContentTransition o
+-- See doc/todo/finish_sync_content_transition.mdwn
+warnSyncContentTransition :: SyncOptions -> [Remote] -> Annex ()
+warnSyncContentTransition o remotes
 	| operationMode o /= SyncMode = noop
 	| isJust (noContentOption o) || isJust (contentOption o) = noop
 	| not (null (contentOfOption o)) = noop
 	| otherwise = getGitConfigVal' annexSyncContent >>= \case
 		HasGlobalConfig (Just _) -> noop
 		HasGitConfig (Just _) -> noop
-		_ -> showwarning
+		_ -> do
+			m <- preferredContentMap
+			hereu <- getUUID
+			when (any (preferredcontentconfigured m) (hereu:map Remote.uuid remotes)) $
+				showwarning
   where
 	showwarning = earlyWarning $
-		"git-annex sync will change default behavior to operate on"
-		<> " --content in a future version of git-annex. Recommend"
-		<> " you explicitly use --no-content (or -g) to prepare for"
-		<> " that change. (Or you can configure annex.synccontent)"
+		"git-annex sync will change default behavior in the future to"
+		<> " send content to repositories that have"
+		<> " preferred content configured. If you do not want this to"
+		<> " send any content, use --no-content (or -g)"
+		<> " to prepare for that change."
+		<> " (Or you can configure annex.synccontent)"
+	preferredcontentconfigured m u = 
+		maybe False (not . isEmpty . fst) (M.lookup u m)
 
 notOnlyAnnex :: SyncOptions -> Annex Bool
 notOnlyAnnex o = not <$> onlyAnnex o
@@ -1135,14 +1160,11 @@ isExport = exportTree . Remote.config
 isImport :: Remote -> Bool
 isImport = importTree . Remote.config
 
+isProxied :: Remote -> Bool
+isProxied = isJust . remoteAnnexProxiedBy . Remote.gitconfig
+
+exportHasAnnexObjects :: Remote -> Bool
+exportHasAnnexObjects = annexObjects . Remote.config
+
 isThirdPartyPopulated :: Remote -> Bool
 isThirdPartyPopulated = Remote.thirdPartyPopulated . Remote.remotetype
-
-splitRemoteAnnexTrackingBranchSubdir :: Git.Ref -> (Git.Ref, Maybe TopFilePath)
-splitRemoteAnnexTrackingBranchSubdir tb = (branch, subdir)
-  where
-	(b, p) = separate' (== (fromIntegral (ord ':'))) (Git.fromRef' tb)
-	branch = Git.Ref b
-	subdir = if S.null p
-		then Nothing
-		else Just (asTopFilePath p)
