@@ -22,7 +22,6 @@ import qualified Data.Map as M
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
 import qualified System.FilePath.ByteString as P
-import Control.Exception
 import Data.Default
 
 import Annex.Common
@@ -59,7 +58,6 @@ import Utility.Tmp
 import Logs.Remote
 import Utility.Gpg
 import Utility.SshHost
-import Utility.Tuple
 import Utility.Directory.Create
 import Messages.Progress
 import Types.ProposedAccepted
@@ -112,7 +110,7 @@ gen baser u rc gc rs = do
 	-- that is now available. Also need to set the gcrypt particiants
 	-- correctly.
 	resetup gcryptid r = do
-		let u' = genUUIDInNameSpace gCryptNameSpace gcryptid
+		let u' = genUUIDInNameSpace gCryptNameSpace (encodeBS gcryptid)
 		v <- M.lookup u' <$> remoteConfigMap
 		case (Git.remoteName baser, v) of
 			(Just remotename, Just rc') -> do
@@ -140,6 +138,7 @@ gen' r u c gc rs = do
 		, name = Git.repoDescribe r
 		, storeKey = storeKeyDummy
 		, retrieveKeyFile = retrieveKeyFileDummy
+		, retrieveKeyFileInOrder = pure True
 		, retrieveKeyFileCheap = Nothing
 		, retrievalSecurityPolicy = RetrievalAllKeysSecure
 		, removeKey = removeKeyDummy
@@ -158,7 +157,7 @@ gen' r u c gc rs = do
 		, readonly = Git.repoIsHttp r
 		, appendonly = False
 		, untrustworthy = False
-		, availability = availabilityCalc r
+		, availability = repoAvail r
 		, remotetype = remote
 		, mkUnavailable = return Nothing
 		, getInfo = gitRepoInfo this
@@ -263,7 +262,7 @@ gCryptSetup _ mu _ c gc = go $ fromProposedAccepted <$> M.lookup gitRepoField c
 		case Git.GCrypt.remoteRepoId g (Just remotename) of
 			Nothing -> giveup "unable to determine gcrypt-id of remote"
 			Just gcryptid -> do
-				let u = genUUIDInNameSpace gCryptNameSpace gcryptid
+				let u = genUUIDInNameSpace gCryptNameSpace (encodeBS gcryptid)
 				if Just u == mu || isNothing mu
 					then do
 						method <- setupRepo gcryptid =<< inRepo (Git.Construct.fromRemoteLocation gitrepo False)
@@ -409,9 +408,9 @@ store' repo r rsyncopts accessmethod
 	storersync = fileStorer $ Remote.Rsync.store rsyncopts
 
 retrieve :: Remote -> Remote.Rsync.RsyncOpts -> AccessMethod -> Retriever
-retrieve r rsyncopts accessmethod k p miv sink = do
+retrieve r rsyncopts accessmethod k p dest miv sink = do
 	repo <- getRepo r
-	retrieve' repo r rsyncopts accessmethod k p miv sink
+	retrieve' repo r rsyncopts accessmethod k p dest miv sink
 
 retrieve' :: Git.Repo -> Remote -> Remote.Rsync.RsyncOpts -> AccessMethod -> Retriever
 retrieve' repo r rsyncopts accessmethod
@@ -432,12 +431,12 @@ retrieve' repo r rsyncopts accessmethod
 	retrieversync = fileRetriever $ Remote.Rsync.retrieve rsyncopts
 
 remove :: Remote -> Remote.Rsync.RsyncOpts -> AccessMethod -> Remover
-remove r rsyncopts accessmethod k = do
+remove r rsyncopts accessmethod proof k = do
 	repo <- getRepo r
-	remove' repo r rsyncopts accessmethod k
+	remove' repo r rsyncopts accessmethod proof k
 
 remove' :: Git.Repo -> Remote -> Remote.Rsync.RsyncOpts -> AccessMethod -> Remover
-remove' repo r rsyncopts accessmethod k
+remove' repo r rsyncopts accessmethod proof k
 	| not $ Git.repoIsUrl repo = guardUsable repo (giveup "cannot access remote") $
 		liftIO $ Remote.Directory.removeDirGeneric True
 			(gCryptTopDir repo)
@@ -446,8 +445,8 @@ remove' repo r rsyncopts accessmethod k
 	| accessmethod == AccessRsyncOverSsh = removersync
 	| otherwise = unsupportedUrl
   where
-	removersync = Remote.Rsync.remove rsyncopts k
-	removeshell = Ssh.dropKey repo k
+	removersync = Remote.Rsync.remove rsyncopts proof k
+	removeshell = Ssh.dropKey repo proof k
 
 checkKey :: Remote -> Remote.Rsync.RsyncOpts -> AccessMethod -> CheckPresent
 checkKey r rsyncopts accessmethod k = do
@@ -489,7 +488,7 @@ toAccessMethod _ = AccessRsyncOverSsh
 getGCryptUUID :: Bool -> Git.Repo -> Annex (Maybe UUID)
 getGCryptUUID fast r = do
 	dummycfg <- liftIO dummyRemoteGitConfig
-	(genUUIDInNameSpace gCryptNameSpace <$>) . fst
+	(genUUIDInNameSpace gCryptNameSpace . encodeBS <$>) . fst
 		<$> getGCryptId fast r dummycfg
 
 coreGCryptId :: ConfigKey
@@ -507,16 +506,25 @@ getGCryptId :: Bool -> Git.Repo -> RemoteGitConfig -> Annex (Maybe Git.GCrypt.GC
 getGCryptId fast r gc
 	| Git.repoIsLocal r || Git.repoIsLocalUnknown r = extract <$>
 		liftIO (catchMaybeIO $ Git.Config.read r)
-	| not fast = extract . liftM fst3 <$> getM (eitherToMaybe <$>)
-		[ Ssh.onRemote NoConsumeStdin r (\f p -> liftIO (Git.Config.fromPipe r f p Git.Config.ConfigList), return (Left $ giveup "configlist failed")) "configlist" [] []
-		, getConfigViaRsync r gc
+	| not fast = extract <$> getM id
+		[ Ssh.onRemote NoConsumeStdin r (configpipe, return Nothing) "configlist" [] []
+		, getconfig $ getConfigViaRsync r gc
 		]
 	| otherwise = return (Nothing, r)
   where
 	extract Nothing = (Nothing, r)
 	extract (Just r') = (fromConfigValue <$> Git.Config.getMaybe coreGCryptId r', r')
 
-getConfigViaRsync :: Git.Repo -> RemoteGitConfig -> Annex (Either SomeException (Git.Repo, S.ByteString, String))
+	configpipe f p = getconfig $ liftIO $ 
+		Git.Config.fromPipe r f p Git.Config.ConfigList
+
+	getconfig a = do
+		(r', _, exitcode, _) <- a
+		if exitcode == ExitSuccess
+			then return (Just r')
+			else return Nothing
+
+getConfigViaRsync :: Git.Repo -> RemoteGitConfig -> Annex (Git.Repo, S.ByteString, ExitCode, String)
 getConfigViaRsync r gc = do
 	let (rsynctransport, rsyncurl, _) = rsyncTransport r gc
 	opts <- rsynctransport
