@@ -5,8 +5,6 @@
  - Licensed under the GNU AGPL version 3 or higher.
  -}
 
-{-# LANGUAGE CPP #-}
-
 module Test.Framework where
 
 import Test.Tasty
@@ -73,7 +71,11 @@ import qualified Command.Uninit
 -- In debug mode, the output is allowed to pass through.
 -- So the output does not get checked in debug mode.
 testProcess :: String -> [String] -> Maybe [(String, String)] -> (Bool -> Bool) -> (String -> Bool) -> String -> Assertion
-testProcess command params environ expectedret expectedtranscript faildesc = do
+testProcess command params environ expectedret expectedtranscript faildesc =
+	void $ testProcess' command params environ expectedret expectedtranscript faildesc
+
+testProcess' :: String -> [String] -> Maybe [(String, String)] -> (Bool -> Bool) -> (String -> Bool) -> String -> IO String
+testProcess' command params environ expectedret expectedtranscript faildesc = do
 	let p = (proc command params) { env = environ }
 	debug <- testDebug . testOptions <$> getTestMode
 	if debug
@@ -81,10 +83,12 @@ testProcess command params environ expectedret expectedtranscript faildesc = do
 			ret <- withCreateProcess p $ \_ _ _ pid ->
 				waitForProcess pid
 			(expectedret (ret == ExitSuccess)) @? (faildesc ++ " failed with unexpected exit code")
+			return ""
 		else do
 			(transcript, ret) <- Utility.Process.Transcript.processTranscript' p Nothing
 			(expectedret ret) @? (faildesc ++ " failed with unexpected exit code (transcript follows)\n" ++ transcript)
 			(expectedtranscript transcript) @? (faildesc ++ " failed with unexpected output (transcript follows)\n" ++ transcript)
+			return transcript
 
 -- Run git. (Do not use to run git-annex as the one being tested
 -- may not be in path.)
@@ -140,12 +144,12 @@ annexeval a = do
 		a `finally` Annex.Action.stopCoProcesses
 
 innewrepo :: IO () -> IO ()
-innewrepo a = withgitrepo $ \r -> indir r a
+innewrepo a = withgitrepo $ \r -> intopdir r a
 
 inmainrepo :: IO a -> IO a
 inmainrepo a = do
 	d <- mainrepodir
-	indir d a
+	intopdir d a
 
 with_ssh_origin :: (Assertion -> Assertion) -> (Assertion -> Assertion)
 with_ssh_origin cloner a = cloner $ do
@@ -160,7 +164,7 @@ with_ssh_origin cloner a = cloner $ do
 	config = "remote.origin.url"
 
 intmpclonerepo :: Assertion -> Assertion
-intmpclonerepo a = withtmpclonerepo $ \r -> indir r a
+intmpclonerepo a = withtmpclonerepo $ \r -> intopdir r a
 
 checkRepo :: Types.Annex a -> FilePath -> IO a
 checkRepo getval d = do
@@ -170,11 +174,11 @@ checkRepo getval d = do
 
 intmpbareclonerepo :: Assertion -> Assertion
 intmpbareclonerepo a = withtmpclonerepo' (newCloneRepoConfig { bareClone = True } ) $
-	\r -> indir r a
+	\r -> intopdir r a
 
 intmpsharedclonerepo :: Assertion -> Assertion
 intmpsharedclonerepo a = withtmpclonerepo' (newCloneRepoConfig { sharedClone = True } ) $
-	\r -> indir r a
+	\r -> intopdir r a
 
 withtmpclonerepo :: (FilePath -> Assertion) -> Assertion
 withtmpclonerepo = withtmpclonerepo' newCloneRepoConfig
@@ -200,14 +204,19 @@ withgitrepo a = do
 	maindir <- mainrepodir
 	bracket (setuprepo maindir) return a
 
-indir :: FilePath -> IO a -> IO a
-indir dir a = do
+intopdir :: FilePath -> IO a -> IO a
+intopdir dir a = do
+	topdir <- Utility.Env.getEnvDefault "TOPDIR" (error "TOPDIR not set")
+	inpath (topdir ++ "/" ++ dir) a
+
+inpath :: FilePath -> IO a -> IO a
+inpath path a = do
 	currdir <- getCurrentDirectory
 	-- Assertion failures throw non-IO errors; catch
 	-- any type of error and change back to currdir before
 	-- rethrowing.
 	r <- bracket_
-		(changeToTopDir dir)
+		(setCurrentDirectory path)
 		(setCurrentDirectory currdir)
 		(tryNonAsync a)
 	case r of
@@ -215,7 +224,7 @@ indir dir a = do
 		Left e -> throwM e
 
 adjustedbranchsupported :: FilePath -> IO Bool
-adjustedbranchsupported repo = indir repo $ Annex.AdjustedBranch.isGitVersionSupported
+adjustedbranchsupported repo = intopdir repo $ Annex.AdjustedBranch.isGitVersionSupported
 
 setuprepo :: FilePath -> IO FilePath
 setuprepo dir = do
@@ -248,7 +257,7 @@ clonerepo old new cfg = do
 		]
 	git "clone" cloneparams "git clone"
 	configrepo new
-	indir new $ do
+	intopdir new $ do
 		ver <- annexVersion <$> getTestMode
 		git_annex "init" 
 			[ "-q"
@@ -257,12 +266,12 @@ clonerepo old new cfg = do
 			]
 			"git annex init"
 	unless (bareClone cfg) $
-		indir new $
+		intopdir new $
 			setupTestMode
 	return new
 
 configrepo :: FilePath -> IO ()
-configrepo dir = indir dir $ do
+configrepo dir = intopdir dir $ do
 	-- ensure git is set up to let commits happen
 	git "config" ["user.name", "Test User"]
 		"git config"
@@ -326,11 +335,7 @@ setTestEnv a = Utility.Tmp.Dir.withTmpDir "testhome" $ \tmphome -> do
 	a
 
 removeDirectoryForCleanup :: FilePath -> IO ()
-#if MIN_VERSION_directory(1,2,7)
 removeDirectoryForCleanup = removePathForcibly
-#else
-removeDirectoryForCleanup = removeDirectoryRecursive
-#endif
 
 cleanup :: FilePath -> IO ()
 cleanup dir = whenM (doesDirectoryExist dir) $ do
@@ -560,11 +565,6 @@ setupTestMode = do
 		git "commit" ["--allow-empty", "-m", "empty"] "git commit failed"
 		git_annex "adjust" ["--unlock"] "git annex adjust failed"
 
-changeToTopDir :: FilePath -> IO ()
-changeToTopDir t = do
-	topdir <- Utility.Env.getEnvDefault "TOPDIR" (error "TOPDIR not set")
-	setCurrentDirectory $ topdir ++ "/" ++ t
-
 tmpdir :: String
 tmpdir = ".t"
 
@@ -691,7 +691,7 @@ origBranch = maybe "foo"
 
 {- Set up repos as remotes of each other. -}
 pair :: FilePath -> FilePath -> Assertion
-pair r1 r2 = forM_ [r1, r2] $ \r -> indir r $ do
+pair r1 r2 = forM_ [r1, r2] $ \r -> intopdir r $ do
 	when (r /= r1) $
 		git "remote" ["add", "r1", "../" ++ r1] "remote add"
 	when (r /= r2) $
@@ -773,7 +773,28 @@ parallelTestRunner' numjobs opts mkts
 				r <- a n
 				worker (r:rs) nvar a
 	
-	go Nothing = withConcurrentOutput $ do
+	summarizeresults a = do
+		starttime <- getCurrentTime
+		(numts, exitcodes) <- a
+		duration <- Utility.HumanTime.durationSince starttime
+		case nub (filter (/= ExitSuccess) (concat exitcodes)) of
+			[] -> do
+				putStrLn ""
+				putStrLn $ "All tests succeeded. (Ran "
+					++ show numts
+					++ " test groups in " 
+					++ Utility.HumanTime.fromDuration duration
+					++ ")"
+				exitSuccess
+			[ExitFailure 1] -> do
+				putStrLn "  (Failures above could be due to a bug in git-annex, or an incompatibility"
+				putStrLn "   with utilities, such as git, installed on this system.)"
+				exitFailure
+			_ -> do
+				putStrLn $ "  Test subprocesses exited with unexpected exit codes: " ++ show (concat exitcodes)
+				exitFailure
+
+	go Nothing = summarizeresults $ withConcurrentOutput $ do
 		ensuredir tmpdir
 		crippledfilesystem <- fst <$> Annex.Init.probeCrippledFileSystem'
 			(toRawFilePath tmpdir)
@@ -801,27 +822,10 @@ parallelTestRunner' numjobs opts mkts
 			(_, _, _, pid) <- createProcessConcurrent p
 			waitForProcess pid
 		nvar <- newTVarIO (1, length ts)
-		starttime <- getCurrentTime
 		exitcodes <- forConcurrently [1..numjobs] $ \_ -> 
 			worker [] nvar runone
 		unless (keepFailuresOption opts) finalCleanup
-		duration <- Utility.HumanTime.durationSince starttime
-		case nub (filter (/= ExitSuccess) (concat exitcodes)) of
-			[] -> do
-				putStrLn ""
-				putStrLn $ "All tests succeeded. (Ran "
-					++ show (length ts) 
-					++ " test groups in " 
-					++ Utility.HumanTime.fromDuration duration
-					++ ")"
-				exitSuccess
-			[ExitFailure 1] -> do
-				putStrLn "  (Failures above could be due to a bug in git-annex, or an incompatibility"
-				putStrLn "   with utilities, such as git, installed on this system.)"
-				exitFailure
-			_ -> do
-				putStrLn $ "  Test subprocesses exited with unexpected exit codes: " ++ show (concat exitcodes)
-				exitFailure
+		return (length ts, exitcodes)
 	go (Just subenvval) = case readish subenvval of
 		Nothing -> error ("Bad " ++ subenv)
 		Just (n, crippledfilesystem, adjustedbranchok) -> setTestEnv $ do
@@ -852,13 +856,7 @@ initTestsName :: String
 initTestsName = "Init Tests"
 
 tastyParser :: [TestTree] -> ([String], Parser Test.Tasty.Options.OptionSet)
-#if MIN_VERSION_tasty(1,3,0)
-tastyParser ts = go
-#else
-tastyParser ts = ([], go)
-#endif
-  where
- 	go = suiteOptionParser ingredients (topLevelTestGroup ts)
+tastyParser ts = suiteOptionParser ingredients (topLevelTestGroup ts)
 
 ingredients :: [Ingredient]
 ingredients =
